#include <string.h>

#include <glib-object.h>

#include "gegl/gegl-types.h"
#include "gegl/gegl-pad.h"

#include "op_filter.h"
#include "canvas/types.h"
#include "canvas/canvas_rect.h"
#include "canvas/canvas.h"

enum
{
  PROP_0,
  PROP_OUTPUT,
  PROP_INPUT,
  PROP_IN_PLACE,
  PROP_LAST
};

static void     op_filter_class_init (OpFilterClass *klass);

static void     op_filter_init       (OpFilter      *self);

static void     finalize             (GObject       *gobject);

static void     get_property         (GObject       *gobject,
                                      guint          prop_id,
                                      GValue        *value,
                                      GParamSpec    *pspec);

static void     set_property         (GObject       *gobject,
                                      guint          prop_id,
                                      const GValue  *value,
                                      GParamSpec    *pspec);

static gboolean evaluate             (GeglOperation *operation,
                                      const gchar   *output_prop);

static void     associate            (GeglOperation *operation);



G_DEFINE_TYPE (OpFilter, op_filter, GEGL_TYPE_OPERATION)


static void
op_filter_class_init (OpFilterClass * klass)
{
  GObjectClass    *object_class = G_OBJECT_CLASS (klass);
  GeglOperationClass *operation_class = GEGL_OPERATION_CLASS (klass);

  object_class->set_property = set_property;
  object_class->get_property = get_property;
  object_class->finalize     = finalize;

  operation_class->evaluate = evaluate;
  operation_class->associate = associate;

  g_object_class_install_property (object_class, PROP_OUTPUT,
                                   g_param_spec_object ("output",
                                                        "Output",
                                                        "An output property",
                                                        CANVAS_TYPE_RECT,
                                                        G_PARAM_READABLE |
                                                        GEGL_PAD_OUTPUT));

  g_object_class_install_property (object_class, PROP_INPUT,
                                   g_param_spec_object ("input",
                                                        "Input",
                                                        "An input property",
                                                        CANVAS_TYPE_RECT,
                                                        G_PARAM_CONSTRUCT |
                                                        G_PARAM_READWRITE |
                                                        GEGL_PAD_INPUT));

  g_object_class_install_property (object_class, PROP_IN_PLACE,
                                   g_param_spec_boolean ("in-place",
                                                        "In place",
                                                        "do processing in-place, modifying the passed in drawable.",
                                                     FALSE,
                                                     G_PARAM_CONSTRUCT |
                                                     G_PARAM_READWRITE |
                                                     GEGL_PAD_INPUT));
}

static void
op_filter_init (OpFilter *self)
{
  self->input  = NULL;
  self->output = NULL;
  self->in_place = FALSE;
}

static void
associate (GeglOperation *self)
{
  GeglOperation  *operation    = GEGL_OPERATION (self);
  GObjectClass   *object_class = G_OBJECT_GET_CLASS (self);

  gegl_operation_create_pad (operation,
                               g_object_class_find_property (object_class,
                                                             "output"));
  gegl_operation_create_pad (operation,
                               g_object_class_find_property (object_class,
                                                             "input"));
}

static void
finalize (GObject *object)
{
  OpFilter *self = OP_FILTER (object);

  if (self->output)
    g_object_unref (self->output);

  G_OBJECT_CLASS (op_filter_parent_class)->finalize (object);
}

static void
get_property (GObject      *object,
              guint         prop_id,
              GValue       *value,
              GParamSpec   *pspec)
{
  OpFilter *self = OP_FILTER (object);

  switch (prop_id)
  {
    case PROP_OUTPUT:
      g_value_set_object (value, self->output);
      break;
    case PROP_INPUT:
      g_value_set_object (value, self->input);
      break;
    case PROP_IN_PLACE:
      g_value_set_boolean (value, self->in_place);
      break;
    default:
      break;
  }
}

static void
set_property (GObject      *object,
              guint         prop_id,
              const GValue *value,
              GParamSpec   *pspec)
{
  OpFilter *self = OP_FILTER (object);

  switch (prop_id)
  {
    case PROP_INPUT:
      self->input = g_value_get_object(value);
      break;
    case PROP_IN_PLACE:
      self->in_place = g_value_get_boolean (value);
      break;
    default:
      break;
  }
}

static gboolean
evaluate (GeglOperation *operation,
          const gchar   *output_prop)
{
  OpFilter      *op_filter = OP_FILTER (operation);
  OpFilterClass *klass     = OP_FILTER_GET_CLASS (operation);
  gboolean success = FALSE;

  if (op_filter->input != NULL)
    {
      success = klass->evaluate (operation, output_prop);
      g_object_unref (op_filter->input);
    }
  else
    {
      g_warning ("%s received NULL input", G_OBJECT_TYPE_NAME (operation));
    }

  return success;
}
